
;+-----------------------------------------------------------+
;!     projekt:   DIN66019 / COMBIVIS                        !
;+-----------------------------------------------------------+
;!     File (Directory):                                     !
;+-----------------------------------------------------------+
;!     Program-/Modulname:  66019.asm                        !
;+-----------------------------------------------------------+
;!     Hardware: PC					     !
;+-----------------------------------------------------------+
;!     Autor: Kaiser,T. 	Devision: Entwicklung	     !
;+-----------------------------------------------------------+
;!     Date (create): 10-Oct-1991			     !
;+-----------------------------------------------------------+
;!     Date:Tue 29-Sep-1992 08:52:28                        !
;+-----------------------------------------------------------+
;!     Version: 3.0                                          !
;+-----------------------------------------------------------+
;!     Construction-Number:				     !
;+-----------------------------------------------------------+
;!	     COPYRIGHT (c) Karl E. Brinkmann GmbH	     !
;+-----------------------------------------------------------+
;!     Files (needed):					     !
;!							     !
;+-----------------------------------------------------------+
;!     Description: ASM-Routines for Protocol                !
;!                  DIN66019 - Interrupt Mode -              !                                !
;!                  for linking with high-level languages    !
;!                  Microsoft QBasic / Microsoft C           !
;!                  Timer funktion 1 ms integrated           !
;+-----------------------------------------------------------+


;To generate code for Microsoft-C use the assembler command line
;     "Masm /DCLANG 66019.asm"
;otherwise code for Microsoft-QBasic will be generated



;DESCRIPTION OF THE PARAMETERS:
;The communication routines use an integer table of 7 words to exchange
;data with the main program. A global variable 'Status' is also used. The
;address of this variable must not be altered in memory while the system
;is in use!
;'Status' is altered by the communication routines or the transmit/receive
;interrupt routines :
;
;  0 =OK-No error, data valid (ACK)
;  >1=Slave-errorcodes (NAK)
;  -1=not used
;  -2=BCC-error
;  -3=invalid start character
;  -4=invalid acknowledge / telegram length
;  -5=framing error
;  -6=overrun error (does not occur on proper systems)
;  -7=parity error
;  -8=unknown / invalid function
;  -9=unknown error
; -10=communication adapter not ready
; -21=acknowledge available - ready for analyzing
; -22=transmitting in progress
; -23=transmitting ready - wait for acknowledge
;
;The words in the integer table have different meanings on different functions,
;the function number is always entered in word 1 :
;
;A.Initialisation of systen (Function 1)
;  Word 0 = Baudrate divided by 10        120..7680   (960 = 9600 Baud)
;  Word 1 = Function number               1
;  Word 2 = Mode of serial adapter        (see below)
;  Word 3 = TTick segment                 segment of integer variable TTICK
;  Word 4 = TTick offset                  offset of variable TTICK (s.below)
;  Word 5 = not used
;  Word 6 = Portaddress of serial adapter (3F8h=COM1 ; 2F8h=COM2)
;    Mode :  Bit 0+1=Data length
;               000XXX00= 5 Bit
;               000XXX01= 6 Bit
;               000XXX10= 7 Bit
;               000XXX11= 8 Bit
;            Bit 2=Number of stop bits
;               000XX0XX= 1 Stoppbit
;               000XX1XX= 1.5/2 Stoppbits
;            Bit 3=Parity Enable
;               000X0XXX= with parity
;               000X1XXX= without parity
;            Bit 4=Parity Even Enable
;               0000XXXX= Parity odd
;               0001XXXX= Parity even
;            Bit 5 to 7 must always be 0
;               000XXXXX
; TTick : Global variable of the typ Word (16bit).This variable is incremented
;         by interrupt every 1 ms. It can be used from main program for time
;         measurement.The address of this variable must not be altered in
;         memory while the system is in use!
;
;
;B.Resetting of used interrupts, IRQ vectors and timers (Function 2)
;  Word 0 = not used
;  Word 1 = Function number               2
;  Word 2 = not used
;  Word 3 = not used
;  Word 4 = not used
;  Word 5 = not used
;  Word 6 = Portaddress of serial adapter (3F8h / 2F8h)
;!!! it is very important to call this function before leaving to DOS !!!
;
;C.Communication functions
;  Word 0 = Slave address (0-255)         only lowbyte is taken
;  Word 1 = Function number               (see below)
;  Word 2 = Parameter address
;  Word 3 = Data
;  Word 4 = Reserved
;  Word 5 = Reserved
;  Word 6 = Portaddress of serial adapter (3F8h / 2F8h)
;  Functions:  11h = DIN66019 Read data
;              12h = DIN66019 Write data
;              13h = DIN66019 Analyze Write acknowledge
;              14h = DIN66019 Analyze Read acknowledge
;



   Code        SEGMENT word Public 'Code'
               Assume CS:Code

   LF          EQU 0Ah
   EOT         EQU 04h
   ENQ         EQU 05h
   STX         EQU 02h
   ETX         EQU 03h
   ACK	       EQU 06h
   NAK	       EQU 15h

Timer_old EQU  08h * 4              ;Timerinterrupt from DOS
PIC1	  EQU  021h		    ;PIC Control PORT
INT_AUS   EQU  0FFh                 ;disable all Interupts

PUBLIC Comadr,sendBuffer,recBuffer,recBufferEnd,LENrecBuffer,wrptr,sendptr
PUBLIC statusseg,statusofs,erreg,More,Sendcnt,Ttickseg,Ttickofs,int08ofs,int08seg
PUBLIC asc_hex,hex_asc,convert_tx_buff,convert_rx_table

ifdef KEBSRC
extrn   T_READ_MOD:PROC,A_READ_MOD:PROC,T_WRITE_MOD:PROC,A_Write_MOD:PROC
extrn   T_READ_KEBI:PROC,A_READ_KEBI:PROC,T_WRITE_KEBI:PROC
endif

;Local Variables--------------------------
 Comadr        DW  3F8h
sendBuffer     DW  100 Dup(?)
recBuffer      DB  100 Dup(?)
recBufferEnd   DB  ?
LENrecBuffer   EQU recbufferend-recbuffer
wrptr          DW  ?
sendptr        DW  ?
statusseg      DW  ?
statusofs      DW  ?
erreg          DB  0
More           DB  0
Sendcnt        DB  0
Ttickseg       DW  ?
Ttickofs       DW  ?
int08ofs       DW  ?
int08seg       DW  ?


;=============================================================================
;Functions for the communication
;Calling for C    : Communicate( int far *Status, int far *Array)
;Calling for Basic: CALLS Communicate( Array(0), status )

ifdef CLANG
   _Communicate PROC far
   Public _Communicate
else
   Communicate PROC far
   Public Communicate
endif
;=============================================================================


OfsStatus      equ [BP+6]
SegStatus      equ [BP+8]
IntOff	       equ [BP+10]
IntSeg	       equ [BP+12]

          Push bp
	  mov  bp,sp
          push ax                       ;save all used registers
	  push bx
	  push dx
	  push es
	  push di
	  push si
;--------- get adress of integer array----------------------------------------
          mov  SI,IntOff                ;load Offset to SI
          mov  ES,IntSeg                ;load Segment to ES
                                        ;ES:SI points always to the beginning
                                        ;of the communication table !!!!

;--------- save adress of variable 'STATUS'-----------------------------------
          mov  ax,segstatus
          mov  cs:statusseg,ax
          mov  ax,ofsstatus
          mov  cs:statusofs,ax

;---------------- get adress of serial adapter--------------------------------
          mov  ax,es:[si+12]            ;load 6th word from Table
          mov  cs:comadr,ax             ;and save it

;---------load function from Table and do the desired task -------------------
          mov  ax,es:[si+2]

sel_1:	  cmp  ax,1
	  jne  sel_2
          jmp  Do_INIT                  ;initialisation of system

sel_2:	  cmp  ax,2
	  jne  sel_3
          jmp  Do_RESET                 ;reset all int-vectors and timer

sel_3:	  cmp  ax,11h
	  jne  sel_4
          call T_READ_66                ;DIN66019 : read data
	  jmp  ser_com_end

sel_4:	  cmp  ax,12h
	  jne  sel_5
          call T_WRITE_66
          jmp  ser_com_end              ;DIN66019 : write data

sel_5:	  cmp  ax,13h
          jne  sel_6
          call A_READ_66                ;DIN66019 : analyze read-acknowledge
	  jmp  ser_com_end

sel_6:	  cmp  ax,14h
          jne  sel_7
          call A_Write_66               ;DIN66019 : analyze write-acknowledge
          jmp  ser_com_end





;=============================================================================
;The following function codes are for KEB internal use only !!
ifdef KEBSRC
sel_7:    cmp  ax,21h
          jne  sel_8
          call T_READ_MOD               ;PARBUS : read data
          jmp  ser_com_end
sel_8:    cmp  ax,22h
          jne  sel_9
          call T_WRITE_MOD              ;PARBUS : write data
          jmp  ser_com_end
sel_9:    cmp  ax,23h
          jne  sel_10
          call A_READ_MOD               ;PARBUS : analyze read-acknowledge
	  jmp  ser_com_end
sel_10:   cmp  ax,24h
          jne  sel_11
          call A_Write_MOD              ;PARBUS : analyze write-acknowledge
          jmp  ser_com_end
sel_11:   cmp  ax,31h
          jne  sel_12
          jmp  Do_T_Read_KEBI           ;KEBI : read data
sel_12:   cmp  ax,32h
          jne  sel_13
          jmp  Do_T_Write_KEBI          ;KEBI : write data
sel_13:   cmp  ax,33h
          jne  sel_14
          jmp  Do_A_Read_KEBI           ;KEBI : analyze read-acknowledge
else
sel_7:
endif
;=============================================================================

sel_14:
          ;no valid function selected - return with errorcode
          mov  ax,-8
          jmp  ser_com_end


;=============================================================================
;Initialisation of timer, serial adapter and interrupt system
Do_INIT:
;=============================================================================
          cli
;-------- save adress of variable 'Ttick' -----------------------------------
          mov  ax,ES:[si+6]             ;get 3rd Word from Table
          mov  cs:Ttickseg,ax           ;      (segment of variable)
          mov  ax,ES:[si+8]             ;also get 4th Word
          mov  cs:Ttickofs,ax           ;      (offset of variable)
;-------- load base adress of serial adapter to DI --------------------------
	  mov  di,ES:[si+12]		;Schnittstellen-Adresse in di
;-------- select desired baud rate -----------------------------------------
          mov  bx,ES:[si]               ;Baudrate/10 in bx
          mov  ax,11520                 ;denomonator
          xor  dx,dx                    ;MSW denominator
          div  bx                       ;result is in AX
          push ax                       ;save result
	  mov  al,80h
          mov  dx,di                    ;put adress of adapter to DX
          add  dx,3                     ;DX = line control register
          out  dx,al                    ;set DLAB
	  pop  ax
	  sub  dx,3
          out  dx,al                    ;LSB Divisor Latch
	  inc  dx
	  mov  al,ah
          out  dx,al                    ;MSB Divisor Latch
;-------- set serial mode ---------------------------------------------------
          mov  ax,ES:[si+4]             ;mode in AX (only al valid)
          mov  dx,di                    ;put adress of adapter to DX
          add  dx,3                     ;DX = line control register
          out  dx,al                    ;set mode
;-------- read in-registers -------------------------------------------------
          mov  dx,di                    ;put adress of adapter to DX
          in   al,dx                    ;read receive buffer
	  add  dx,2
          in   al,dx                    ;interrupt identification register
	  add  dx,3
          in   al,dx                    ;line status register
	  inc  dx
          in   al,dx                    ;modem status register
;-------- set hardware interrupt of the 8250 chip ---------------------------
          mov  dx,di                    ;put adress of adapter to DX
          inc  dx                       ;interrupt enable register
          mov  al,0
          out  dx,al                    ;disable all interrupts
          add  dx,3                     ;modem control register
          in   al,dx                    ;read value
          or   al,8
          out  dx,al                    ;switch on OUT2

;-------- set pointer to the begin of receive buffer ------------------------
          mov  cs:wrptr,OFFSET recbuffer

;-------- set INTvektor to the serial interrupt service routine ---------------------
          mov  dx,di
          mov  bl,0Bh                   ;INT-Nr for Com2
          mov  bh,11110111b             ;PIC-Mask IRQ3
          cmp  dx,2F8h
          je   intcom1                  ;Com2 selected
          mov  bl,0Ch                   ;INT-Nr for Com1
          mov  bh,11101111b             ;PIC-Mask IRQ4
intcom1:  push bx                       ;save PIC-Mask
          push ds
          mov  ax,cs
          mov  ds,ax                    ;ds points to this code segment
          mov  dx,OFFSET Read_char      ;dx points to the interrupt routine
          mov  ah,25h                   ;DOS function 25h
          mov  al,bl                    ;Interrupt-nr in al
          int  21h
          pop  ds

;-------- disable all interrupts in PIC -------------------------------------
          pop  bx                       ;load mask to enable COMn
          in   al,PIC1                  ;load previous mask
          and  al,bh                    ;and reset the new bit
          push ax                       ;save the new mask for enabling later
          mov  al,INT_AUS
          out  (PIC1),al                ;disable ALL IRQ's now

;-------- save old INTvector 08 ----------------------------------------------
          push ds
          mov  ax,0
          mov  ds,ax
          mov  bx,Timer_old
          mov  ax,[bx]                  ;get offset of INT 8
          mov  WORD PTR cs:int08ofs,ax  ;and save it to memory
          mov  ax,[bx+2]                ;get also segment of INT 8
          mov  WORD PTR cs:int08seg,ax  ;and save
          pop  ds

;-------- redirect the user timer-Interrupt 08h to the routine  'Incttick'
          push ds
          mov  ax,cs
          mov  ds,ax                    ;DS points to this code segment
          mov  dx,OFFSET Incttick       ;load DX with the offset
          mov  ah,25h                   ;DOS function 25h
          mov  al,08h                   ;INT-No
          int  21h
          pop  ds

;-------- change timer 0 (SYSclock) to 1ms time ------------------------------
          mov  dx,43h                   ;control word Timer 0
          mov  al,036h                  ;mode adjustment Timer 0
          out  dx,al
          mov  dx,40h                   ;R/W-Port Timer 0
          mov  al,0A9h
          out  dx,al                    ;write Divisor LSB
          mov  al,04h
          out  dx,al                    ;write Divisor MSB
                                        ;1193000 / 1193 = 1000 Hz ->1ms rate

;-------- enable PIC and 8250 ------------------------------------------------
          pop  ax                       ;get mask for PIC
          out  (PIC1),al                ;and enable the desired IRQs

          mov  dx,di                    ;reload adress of serial adapter
          in   al,dx                    ;reset any pending interrupt

          mov  dx,020h
          mov  al,020h
          out  dx,al                    ;normal end of interrupt for PIC
          mov  ax,0
          sti
          jmp  ser_com_end


Do_RESET:
;-------- set timer 0 (SYSclock) back to 55ms time period --------------------
          cli
          mov  dx,43h                   ;control word Timer 0
          mov  al,036h                  ;mode adjustment Timer 0
          out  dx,al
          mov  dx,40h                   ;R/W-Port Timer 0
          mov  al,0FFh
          out  dx,al                    ;write Divisor LSB
          out  dx,al                    ;write Divisor MSB
                                        ;1193000/65535 = 18.2 Hz ->55ms rate

;-------- set back old INT 8 -------------------------------------------------
          push ds
          mov  ax,cs:int08seg
          mov  ds,ax                    ;DS points to segment of old INTroutine
          mov  dx,cs:int08ofs           ;DX points to offset of old INTroutine
          mov  ah,25h                   ;DOS funktion 25h
          mov  al,08h                   ;INT-No
          int  21h
          pop  ds

;-------- disable serial interrupt at the PIC --------------------------------
          mov  dx,cs:comadr
          mov  bh,00001000b             ;PIC mask IRQ3
          cmp  dx,2F8h
          je   rescom1                  ;Com2 selected
          mov  bh,00010000b             ;PIC maske IRQ4
rescom1:  mov  dx,21h                   ;port adress PIC1 (IMR)
          in   al,dx                    ;read old mask
          or   al,bh                    ;set appropriate bit
          out  dx,al                    ;disable IRQ3/4
          mov  ax,0
          sti
          jmp  ser_com_end



;=============================================================================
;The following function codes are for KEB internal use only !!
ifdef KEBSRC
Do_T_Read_KEBI:
          mov  ax,-9                    ;preload error code
          mov  bx,es:[si+4]             ;Length/P-Addr
          cmp  bh,0                     ;check length..
          je   ser_com_end
          cmp  bh,3
          ja   ser_com_end
          call T_READ_KEBI
          jmp  ser_com_end

Do_A_Read_KEBI:
          call A_READ_KEBI
          jmp  ser_com_end

Do_T_WRITE_KEBI:
          mov  ax,-9                    ;preload error code
          mov  bx,es:[si+4]             ;Length/P-Addr
          cmp  bh,0                     ;check length..
          je   ser_com_end
          cmp  bh,3
          ja   ser_com_end
          call T_WRITE_KEBI
          jmp  ser_com_end
endif
;=============================================================================



;=============================================================================
; end of the communication routines - reload status ...
;=============================================================================
ser_com_end:
          cmp  ax,-22                   ;Status already set ..
          je  ser_com_skip
          mov  ES,segstatus             ;load segment address
          mov  bx,ofsstatus             ;and offset address
          mov  ES:[bx],ax               ;put AX to this integer variable
ser_com_skip:
	  pop  si
	  pop  di
	  pop  es
	  pop  dx
	  pop  bx
	  pop  ax
	  mov  sp,bp
	  pop  bp
ifdef CLANG
	  ret
_Communicate ENDP
else
	  ret  8
 Communicate ENDP
endif



;#############################################################################
;66019  :  read data
;Uses:
;ES:SI = Points to integer table
;AX    = return code
;#############################################################################
T_READ_66     PROC NEAR
;-------------------------TEST IF BROADCAST ----------------------------
          mov  ax,-8                    ;preload error code
          mov  bx,0                     ;bx=offset table (slave-addr)
          mov  cx,es:[si+bx]            ;load slave-no
          cmp  cl,0F0h                  ;Broadcast-Address ?
          jb   T_READ_66_0              ;adresse valid
          jmp  T_READ_66_end            ;function not possible
;--------------------------FILL SEND BUFFER-----------------------------
T_READ_66_0:
          mov  di,OFFSET sendBuffer
;--------------------------enter EOT --------------------------------------
	  mov  BYTE PTR cs:[di],EOT
          inc  di                       ;increment buffer ptr
;----------------------enter Slave Address------------------------------
          mov  bx,es:[si]               ;load Slave Address from table
          mov  al,bl                    ;save lower nibble to al
	  mov  cl,4
          shr  al,cl                    ;shift upper nibble to lower
          call hex_asc                  ;convert to ascii character
          mov  cs:[di],al               ;enter first Slaveaddress character
          inc  di
          mov  al,bl                    ;reload lower nibble of slave address
          mov  bl,0Fh
          and  al,bl                    ;mask upper nibble
          call hex_asc                  ;convert to ascii character
          mov  cs:[di],al               ;enter second Slaveaddress character
          inc  di
;----------------enter Parameter Address---------------------------
          mov  cx,1                     ;convert 1 Word = 4 characters
          mov  bx,4                     ;BX points to 2nd data word MSB
          call Convert_TX_buff
;-----------------------------enter ENQ ---------------------------------
	  mov  BYTE PTR cs:[di],ENQ
          inc  di
;-----------------set receive pointer to begin of recbuffer----------------
          mov  cs:wrptr,OFFSET recbuffer
          mov  cs:more,11               ;receive max. 11 characters
;-------------------adjust send buffer pointer-----------------------------
          mov  di,OFFSET sendBuffer     ;DI points to sendbuffer
          mov  cs:sendptr,di            ;save this pointer
          mov  cs:sendcnt,8             ;and set the count of characters to send
;--------------switch on transmitter (RTS) ------------------------
          mov  dx,cs:comadr
          add  dx,4                     ;Modem-Control Register
          in   al,dx
          and  al,29                    ;mask for RTS = 0
          out  dx,al                    ;reset RTS
;--------------Set status to transmitting in progress---------------------
          mov  ES,cs:statusseg          ;load segment address
          mov  BX,cs:statusofs          ;and offset address
          mov  WORD PTR ES:[bx],-22     ;Set status to transmitting running
;--------------switch interrupts of 8250 ------------------------
          sub  dx,3                     ;Interrupt enable register
          mov  al,2                     ;disable receive interrupt
          out  dx,al                    ;enable transmitt interrupt
          mov  ax,-22                   ;function code = transmitting running
T_READ_66_end:
          ret
T_READ_66  ENDP


;#############################################################################
;66019  : analyze read acknowledge
;Uses:
;ES:SI = Points to integer table
;AX    = return code
;#############################################################################
 A_READ_66 PROC near
          mov  ah,cs:erreg              ;load errorcode from INT-routine
          mov  al,ah                    ;save to AL
          and  al,00000010b             ;Overrun bit
          cmp  al,0
          je   A_READ_66_1              ;no Overrun
          mov  ax,-6                    ;set return value
          jmp  A_READ_66_end            ;and return
A_READ_66_1:
          mov  al,ah                    ;reload errorcode from INT routine
          and  al,00000100b             ;Parity bit
          cmp  al,0
          je   A_READ_66_2              ;no Parity error
          mov  ax,-7                    ;set return value
          jmp  A_READ_66_end            ;and return
A_READ_66_2:
          mov  al,ah                    ;reload errorcode from INT routine
          and  al,00001000b             ;Framing bit
          cmp  al,0
          je   A_READ_66_3              ;No Framing error
          mov  ax,-5                    ;set return value
          jmp  A_READ_66_end            ;and return
A_READ_66_3:
;-------------------ANALYZE BUFFER -----------------------------------------
;---------------search for first valid character------------------------------
          mov  ax,-9                    ;preload errorcode
          mov  di,OFFSET recBuffer      ;DI points to begin of recBuffer
          mov  cx,LENrecBuffer          ;CX shows len of buffer
get_read_loop:
          mov  bl,al                    ;save last character
          mov  al,CS:[di]               ;load character
          inc  di
          cmp  al,STX
          je   A_READ_66_6              ;STX is valid
          cmp  al,EOT
          je   A_READ_66_4              ;EOT is valid
          loop get_read_loop            ;proceed searching
          mov  ax,-3                    ;no valid character found
          jmp  A_READ_66_end            ;return..
;-----------------------EC EOT found-----------------------------------------
A_READ_66_4:
          xor  ax,ax                     ;clear AX
          mov  al,bl                     ;get character EC
          call asc_hex                   ;convert to hex
          jmp  A_READ_66_end             ;and return
;-------------------BCC CHECK after STX was found ---------------------------
A_READ_66_6:
          call test_bcc                  ;check if BCC is correct
          cmp  ax,0
          je   A_READ_66_7               ;no BCC error
          mov  ax,-2                     ;set return value to BCC error
          jmp  A_READ_66_end             ;and return
;-------------------load parameter address and data to integer table------
A_READ_66_7:
                                        ;DI points to 2nd character in buffer
          mov  bx,4                     ;BX points to 2nd word of integer table
          mov  cx,2                     ;convert 2 words
          call convert_rx_Table
          xor  ax,ax                    ;no error occured
A_READ_66_end:
          mov  BYTE PTR cs:erreg,0      ;reset error variable
          ret                           ;and return
A_READ_66 ENDP


;#############################################################################
;66019  :  write data
;Uses:
;ES:SI = Points to integer table
;AX    = return code
;#############################################################################
T_WRITE_66 PROC NEAR
;--------------------------FILL SEND BUFFER-----------------------------
          mov  di,OFFSET sendBuffer
;--------------------------enter EOT --------------------------------------
	  mov  BYTE PTR cs:[di],EOT
          inc  di                       ;increment buffer ptr
;----------------------enter Slave Address------------------------------
          mov  bx,es:[si]               ;load Slave Address from table
          mov  al,bl                    ;save lower nibble to al
	  mov  cl,4
          shr  al,cl                    ;shift upper nibble to lower
          call hex_asc                  ;convert to ascii character
          mov  cs:[di],al               ;enter first Slaveaddress character
          inc  di
          mov  al,bl                    ;reload lower nibble of slave address
          mov  bl,0Fh
          and  al,bl                    ;mask upper nibble
          call hex_asc                  ;convert to ascii character
          mov  cs:[di],al               ;enter second Slaveaddress character
          inc  di
;--------------------enter start character  ---------------------------
	  mov  BYTE PTR cs:[di],STX
	  inc  di
;----------enter parameter address and data---------------------
          mov  cx,2                     ;convert 2 Word = 8 characters
          mov  bx,4                     ;BX points to 2nd data word MSB
	  call Convert_TX_buff
;-------------enter stop character ---------------------------
	  mov  BYTE PTR cs:[di],ETX
	  inc  di
;-------------------calculate BCC-Code ---------------------------
          mov  di,OFFSET sendbuffer     ;DI -> begin send buffer
          add  di,4                     ;DI -> character after STX
          call make_BCC
;-----------------set receive pointer to begin of recbuffer----------------
          mov  cs:wrptr,OFFSET recbuffer
          mov  cs:more,2                ;receive max. 2 characters
;-------------------adjust send buffer pointer-----------------------------
          mov  di,OFFSET sendBuffer     ;DI points to sendbuffer
          mov  cs:sendptr,di            ;save this pointer
          mov  cs:sendcnt,14            ;and set the count of characters to send
;--------------switch on transmitter (RTS) ------------------------
          mov  dx,cs:comadr
          add  dx,4                     ;Modem-Control Register
          in   al,dx
          and  al,29                    ;mask for RTS = 0
          out  dx,al                    ;reset RTS
;--------------Set status to transmitting in progress---------------------
          mov  ES,cs:statusseg          ;load segment address
          mov  BX,cs:statusofs          ;and offset address
          mov  WORD PTR ES:[bx],-22     ;Set status to transmitting running
;--------------switch interrupts of 8250 ------------------------
          sub  dx,3                     ;Interrupt enable register
          mov  al,2                     ;disable receive interrupt
          out  dx,al                    ;enable transmitt interrupt
          mov  ax,-22                   ;function code = transmitting running
          ret
T_WRITE_66  ENDP


;#############################################################################
;66019  : analyze write acknowlegde
;Uses:
;ES:SI = Points to integer table
;AX    = return code
;#############################################################################
A_WRITE_66  PROC NEAR
          mov  ah,cs:erreg              ;load errorcode from INT-routine
          mov  al,ah                    ;save to AL
          and  al,00000010b             ;Overrun bit
          cmp  al,0
          je   A_WRITE_66_1             ;no Overrun
          mov  ax,-6                    ;set return value
          jmp  A_WRITE_66_end           ;and return
A_WRITE_66_1:
          mov  al,ah                    ;reload errorcode from INT routine
          and  al,00000100b             ;Parity bit
          cmp  al,0
          je   A_WRITE_66_2             ;no Parity error
          mov  ax,-7                    ;set return value
          jmp  A_WRITE_66_end           ;and return
A_WRITE_66_2:
          mov  al,ah                    ;reload errorcode from INT routine
          and  al,00001000b             ;Framing bit
          cmp  al,0
          je   A_WRITE_66_3             ;No Framing error
          mov  ax,-5                    ;set return value
          jmp  A_WRITE_66_end           ;and return
A_WRITE_66_3:
;-------------------ANALYZE BUFFER -----------------------------------------
;---------------search for first valid character------------------------------
          mov  ax,-9                    ;preload error code
          mov  di,OFFSET recBuffer      ;DI points to begin of recBuffer
          mov  cx,LENrecBuffer          ;CX shows len of buffer
A_Write_66_loop:
          mov  bl,al                    ;save last character
          mov  al,CS:[di]               ;load character
          inc  di
          cmp  al,ACK
          je   A_WRITE_66_OK            ;ACK is valid
          cmp  al,NAK
          je   A_WRITE_66_7             ;NAK is valid
          loop A_Write_66_loop          ;proceed searching
          mov  ax,-4                    ;no valid character found
          jmp  A_WRITE_66_end           ;return..
A_WRITE_66_7:
          xor  ax,ax                    ;clear AX
          mov  al,bl                    ;get character EC
          call asc_hex                  ;convert to hex
          jmp  A_WRITE_66_end           ;and return
A_WRITE_66_ok:
          xor  ax,ax                    ;no error occured
A_WRITE_66_end:
          mov  BYTE PTR cs:erreg,0      ;reset error variable
          ret                           ;and return
A_WRITE_66  ENDP



;###########################################################################
; SUBROUTINES
;###########################################################################

;=======================================================
 Make_BCC PROC near
; the BCC code of transmitt buffer is generated and appended
; Uses :
; CS:Buffer = transmitt buffer
;=======================================================
	  push dx
	  push cx
          push di
          xor  dx,dx                    ;clear BCC count register
          mov  cx,9                     ;number of characters in CX
Make_BCC_loop:
          mov  al,cs:[di]               ;get one character
          inc  di                       ;increment buffer ptr
          xor  dl,al                    ;xor with dl
          loop Make_BCC_loop            ;loop for all characters
          cmp  dl,1Fh                   ;BCC less than 20h ?
          ja   Make_BCC_1               ;no
          add  dl,20h                   ;add offset
Make_BCC_1:
          mov  cs:[di],dl               ;and append to buffer
	  pop  di
	  pop  cx
	  pop  dx
	  ret
Make_BCC  ENDP


;=======================================================
 Test_BCC PROC near
; check BCC code of received characters
; Uses :
; CS:recBuffer = receiving buffer
; AX = return code
; DI points to 1st character in buffer
;=======================================================
	  push dx
	  push cx
	  push di
          xor  dx,dx                    ;clear BCC count register
          mov  cx,9                     ;no of characters
test_BCC_loop:
          mov  al,cs:[di]               ;get one character
          inc  di                       ;increment buffer ptr
          cmp  di,OFFSET recBufferEnd   ;end of buffer reached?
          jne  test_bcc_0
          mov  di,OFFSET recbuffer      ;set to beginn of buffer
test_bcc_0:
          xor  dl,al                    ;xor with dl
          loop test_BCC_loop            ;loop for all characters
	  cmp  dl,20h
          ja   test_bcc_1               ;check if offset is needed
          add  dl,20h                   ;add offset
test_bcc_1:
          mov  al,cs:[di]               ;get last character of buffer (BCC)
          cmp  al,dl                    ;and compare with calculated value
          je   test_BCC_ok              ;BCC code ok
          mov  ax,-2                    ;not ok - load error code
          jmp  test_BCC_end             ;and return
test_BCC_ok:
          xor  ax,ax                    ;clear error register
test_BCC_end:
	  pop  di
	  pop  cx
	  pop  dx
	  ret
test_BCC  ENDP


;=======================================================
 convert_rx_Table PROC near
; gets characters from buffer, converts them into words
; and put it into integer table
; Uses :
; Asc_hex (Proc)
; CS:Buffer = receive buffer
; Di=points to next character in receive buffer (is incremented)
; ES:SI=Start address of integer table
; BX=Offset in table (is incremented)
; CX=no. of words to convert (4 characters = 1 Word)
;=======================================================
	  push ax
	  push dx
convert_rx_Table_begin:
          mov  al,cs:[di]               ;get 1st character
          call asc_hex                  ;convert to hex
	  push cx
          mov  cl,4
          shl  al,cl                    ;shift into upper nibble
	  pop  cx
          mov  ah,al                    ;and save to AH
          inc  di
          mov  al,cs:[di]               ;get 2nd character
          call asc_hex                  ;convert to hex
          or   ah,al                    ;add to byte
          mov  dh,ah                    ;and save as MSB
          inc  di
          mov  al,cs:[di]               ;get 3rd character
          call asc_hex                  ;convert to hex
          push cx
          mov  cl,4
          shl  al,cl                    ;shift into upper nibble
          pop  cx
          mov  ah,al                    ;and save to AH
          inc  di
          mov  al,cs:[di]               ;get 4th character
          call asc_hex                  ;convert to hex
          or   ah,al                    ;add to byte
          mov  dl,ah                    ;and save as LSB
          mov  es:[si+bx],dx            ;put word into table
          inc  di
          inc  bx                       ;increment table ptr
          inc  bx                       ;increment table ptr
          loop convert_rx_Table_begin   ;proceed for all words
          pop  dx
          pop  ax
          ret
convert_rx_Table ENDP



;=======================================================
 Convert_TX_buff PROC near
; gets data words from integer table,converts them into charcaters
; and puts them into transmitt buffer
; Uses:
; Hex_asc (Proc)
; CS:sendBuffer = transmitt buffer
; Di=points to next character in transmitt buffer (is incremented)
; ES:SI=Start address of integer table
; BX=Offset in table (is incremented)
; CX=no. of words to convert (4 characters = 1 Word)
;=======================================================
          push ax
          push dx
Convert_TX_buff_begin:
          mov  dx,es:[si+bx]            ;get data (1 Word)
          xchg dh,dl                    ;change MSB / LSB
          mov  al,dl                    ;1st nibble
          push cx
          mov  cl,4
          shr  al,cl                    ;shift upper nibble to lower
          pop  cx
          call hex_asc                  ;convert to ascii character
          mov  cs:[di],al               ;and save to buffer
          inc  di
          mov  al,dl                    ;2nd nibble
          and  al,0Fh                   ;clear upper nibble
          call hex_asc                  ;convert to ascii character
          mov  cs:[di],al               ;and save to buffer
          inc  di
          mov  dl,dh                    ;get LSB of data
          mov  al,dl                    ;3rd nibble
          push cx
          mov  cl,4
          shr  al,cl                    ;shift upper nibble to lower
          pop  cx
          call hex_asc                  ;convert to ascii character
          mov  cs:[di],al               ;and save to buffer
          inc  di
          mov  al,dl                    ;4th nibble
          and  al,0Fh                   ;clear upper nibble
          call hex_asc                  ;convert to ascii character
          mov  cs:[di],al               ;and save to buffer
          inc  di                       ;increment buffer ptr
          inc  bx                       ;increment table ptr
          inc  bx                       ;increment table ptr
          loop Convert_TX_buff_begin    ;proceed for all words
          pop  dx
          pop  ax
          ret
Convert_TX_buff ENDP

;=======================================================
   asc_hex PROC near
; converts characters in AL (O-F) to hex value (lower nibble of AL)
;=======================================================
          sub  al,30h
          cmp  al,0Ah
          jb   asc_hex_end
          sub  al,7
asc_hex_end:
          ret
  asc_hex ENDP

;=======================================================
   hex_asc PROC near
; converts lower nibble of AL to ASCII character
;=======================================================
          add  al,30h
          cmp  al,3Ah
          jb   hex_asc_end
          add  al,7
hex_asc_end:
          ret
  hex_asc ENDP



;=======================================================
 Read_Char PROC far
; This is the serial interrupt service routine
; for receiving and transmitting characters
;=======================================================
          cli
          push es
          push dx
          push cx
          push bx
          push ax
          mov  dx,cs:comadr             ;DX=base address of serial adapter
          add  dx,2                     ;Interrupt identification register
          in   al,dx                    ;read
          and  al,7                     ;and mask
          cmp  al,2                     ;THR empty ?
          jne  read_char_begin          ;no - receive one character
          jmp  send_char                ;yes - send next character

;-----------------------------  RECEIVE ----------------------------------
read_char_begin:
          mov  cl,cs:more               ;load max. no of chars to receive
          mov  dx,cs:comadr             ;DX=base address of serial adapter
          in   al,dx                    ;read character
          mov  bx,cs:wrptr              ;load ptr to receive buffer
          mov  cs:[bx],al               ;save character
          inc  bx
          cmp  bx,OFFSET recBufferEnd   ;end of buffer reached ?
          jne  read_char_1              ;no
          mov  bx,OFFSET recbuffer      ;set ptr to begin of buffer
read_char_1:
          mov  cs:wrptr,bx              ;store receive buffer pointer
          mov  ah,al                    ;save character
          add  dx,5                     ;DX=Line status register
          in   al,dx                    ;read
          and  al,00001110b             ;mask error bits
          cmp  al,0
          je   read_char_4
          mov  cs:erreg,al              ;save errorcode
read_char_4:
          cmp  ah,LF
          je   Read_char_term           ;LF=last character (PARBUS)
          cmp  ah,EOT
          je   Read_char_term           ;EOT=last character (66019)
          cmp  ah,ACK
          je   Read_char_term           ;ACK=last character (66019)
          cmp  ah,NAK
          je   Read_char_term           ;NAK=last character (66019)
          cmp  ah,ETX
          jne  Read_char_2              ;not ETX
          mov  cl,2                     ;also read character after ETX (66019)
read_char_2:
          dec  cl
          mov  cs:more,cl               ;store no of characters to receive
          cmp  cl,0
          jne  read_char_end            ;futher characters to receive !

Read_char_term:                         ;no more characters to receive
          mov  es,cs:statusseg
          mov  bx,cs:statusofs
          mov  WORD PTR es:[bx],-21     ;set STATUS to 'got acknowledge'
          jmp  read_char_end

;-----------------------------  TRANSMITT ----------------------------------
send_char:
          cmp  cs:sendcnt,0             ;last character sent ?
          je   wait_for_last            ;yes - wait until transmitter empty
          sub  dx,2                     ;DX=Transmitter holding register
          mov  bx,cs:sendptr            ;load ptr to transmit buffer
          mov  al,cs:[bx]               ;get character
          out  dx,al                    ;and put into THR
          inc  cs:sendptr               ;correct ptr
          dec  cs:sendcnt               ;and no of chararters to transmit
          jmp  read_char_end
wait_for_last:
          add  dx,3                     ;DX=Line status register
send_buff_wait2:
          in   al,dx                    ;read
          and  al,01000000b             ;select TEMT
          cmp  al,01000000b             ;Transmitter empty ?
          jne  send_buff_wait2          ;no - wait
send_buff_end:
          mov  dx,cs:comadr             ;DX=base address of serial adapter
          in   al,dx                    ;clear receiver register
          add  dx,4                     ;Modem control register
          in   al,dx                    ;read
          or   al,2
          out  dx,al                    ;set RTS to 1
          add  dx,1                     ;Line status register
          in   al,dx                    ;clear interrupts and errors
          sub  dx,4                     ;interrupt enable register
          mov  al,1                     ;enable receive interrupt
          out  dx,al                    ;disable transmit interrupt

          mov  es,cs:statusseg
          mov  bx,cs:statusofs
          mov  WORD PTR es:[bx],-23     ;set STATUS to transmitting done

read_char_end:
          mov  dx,020h
          mov  al,020h
          out  dx,al                     ;Normal end of INT

          pop  ax
          pop  bx
          pop  cx
          pop  dx
          pop  es
          sti
          iret

Read_char ENDP

;=======================================================
 IncTtick PROC far
; Increments variable TTICK every 1 ms
;=======================================================
          cli
          push ax
          push es
          push bx
          jmp inc_1
 T0cnt    DB  0                     ;Byte for counting of cycles Timer-0

inc_1:
          mov  es,cs:Ttickseg       ;segment variable ttick
          mov  bx,cs:Ttickofs       ;offset variable ttick
          inc  WORD PTR es:[bx]     ;increment

          mov  al,cs:T0cnt          ;load cycle counter
          inc  al                   ;and increment
          mov  cs:T0cnt,al          ;save value
          cmp  al,55
          jb   no_t0_int            ;no 55 ms left

          ;call original INT 08h
          mov  al,0
          mov  cs:T0cnt,al          ;reset value
          pop  bx
          pop  es
          pop  ax
          pushf                     ;simulate INT
          call DWORD PTR cs:[int08ofs] ;call BIOS function
          sti
          iret                      ;and return

no_t0_int:
          pop  bx
          pop  es
          mov  al,20h               ;Reset 8259 IRQ-Controller
          out  20h,al
          pop  ax
          sti
          iret
 IncTtick Endp


   Code    ENDS
   END
